"""
WebDebug Message List

Keeps track of all HTTP messages received by WebDebug, generates statistics for them
and renders them to HTML.


WebDebug is a web development proxy server that captures and displays HTTP
requests and responses.
Copyright (C) 1999  Paul Clip.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""

### IMPORTS

from WDconst import *
import WDconfig
import HTTPrequest
import HTTPresponse

if not WDconfig.SINGLE_THREADED:
	import threading
import time


### CONSTANTS

# indexes for statistics gathering
STAT_NUMB = 0
STAT_LEN = 1

# indexes for saving and restoring state
MLIST = "messagelist"
FIRSTTIME = "firsttime"
LASTTIME = "lasttime"
WDVERSION = "WDversion"


### Classes

class MessageList:
	"""
	Keeps track of HTTP requests and responses belonging to a single session.
	Member variables:
		messagelist: a (chronological) sequence of HTTP request/response pairs
		listlock: a lock protecting the messagelist from simultaneous access
		firstmessagetime: the time at which the first message pair was added
		lastmessagetime: the time at which the last message pair was added
		sortorder: the order in which the message pairs should be sorted (default is request received order)
		inorder: denotes whether the message list is sorted according to sort order or not
		WDversion: the version of WebDebug this messagelist was created in
	"""

	def __init__(self):
		self.messagelist = []
		if not WDconfig.SINGLE_THREADED:
			self.listlock = threading.Lock()
		self.firstmessagetime = 0
		self.lastmessagetime = 0
		self.sortorder = REQUEST_SORT_ORDER
		self.inorder = NO
		self.WDversion = VERSION_NUMBER

	def __getstate__(self):
		""" returns the state to pickle """
		return { MLIST: self.messagelist,
			 FIRSTTIME: self.firstmessagetime,
			 LASTTIME: self.lastmessagetime,
			 WDVERSION: self.WDversion }

	def __setstate__(self, state):
		""" sets the state when unpickling """
		self.__init__()
		self.messagelist = state[MLIST]
		self.firstmessagetime = state[FIRSTTIME]
		self.lastmessagetime = state[LASTTIME]
		self.WDversion = state[WDVERSION]

	def lockList(self):
		if not WDconfig.SINGLE_THREADED:
			self.listlock.acquire()

	def unlockList(self):
		if not WDconfig.SINGLE_THREADED:
			self.listlock.release()

	def addMessagePair(self, request, response):
		""" Add a request/response pair to the message list """
		pair = { REQUEST: request, RESPONSE: response }
		if self.firstmessagetime == 0 or self.firstmessagetime > request.getLoadTime():
			self.firstmessagetime = request.getLoadTime()
		if self.lastmessagetime == 0 or self.lastmessagetime < response.getLoadTime():
			self.lastmessagetime = response.getLoadTime()
		self.lockList()
		self.messagelist.append(pair)
		self.unlockList()
		self.inorder = NO

	def addRequest(self, request):
		""" Adds a request to the message list, response is set to None """
		pair = { REQUEST: request, RESPONSE: None }
		if self.firstmessagetime == 0 or self.firstmessagetime > request.getLoadTime():
			self.firstmessagetime = request.getLoadTime()
		self.lockList()
		self.messagelist.append(pair)
		self.unlockList()
		self.inorder = NO

	def addResponse(self, request, response):
		""" Adds a response to its corresponding request """
		if self.lastmessagetime == 0 or self.lastmessagetime < response.getLoadTime():
			self.lastmessagetime = response.getLoadTime()
		p = self.findMessagePair(id(request))
		if p[RESPONSE] == None:
			self.lockList()
			p[RESPONSE] = response
			self.unlockList()
		else:
			raise "MessageList.addResponse: error trying to replace existing response"

	def clearList(self):
		""" Clears the message list of its contents """
		self.lockList()
		self.messagelist = []
		self.firstmessagetime = 0
		self.lastmessagetime = 0
		self.unlockList()
		self.inorder = NO

	def sortOn(self, sort_order):
		""" Checks whether to sort the message list on the specified sort order """
		if sort_order == CURRENT_SORT_ORDER:
			if self.inorder:
				return
			else:
				self.sort(self.sortorder)
		elif sort_order != self.sortorder or not self.inorder:
			self.sort(sort_order)

	def sort(self, sort_order):
		""" Sorts the message list on the specified sort order """
		if sort_order == REQUEST_SORT_ORDER:
			self.messagelist.sort(self.requestSortOrder)
		elif sort_order == RESPONSE_SORT_ORDER:
			self.messagelist.sort(self.responseSortOrder)
		elif sort_order == SOCKET_SORT_ORDER:
			self.messagelist.sort(self.socketSortOrder)
		else:
			raise "MessageList.sort(): illegal sort order: %s" % sort_order
		self.sortorder = sort_order
		self.inorder = YES

	def requestSortOrder(self, p1, p2):
		""" Returns -1, 0, or 1 depending on whether p1's request was received earlier, equal to or later than p2's """
		t1 = p1[REQUEST].getCreationTime()
		t2 = p2[REQUEST].getCreationTime()
		if t1 < t2:
			return -1
		elif t1 == t2:
			return 0
		else:
			return 1

	def responseSortOrder(self, p1, p2):
		""" Returns -1, 0, or - depending on whether p1's response was downloaded earlier, equal to or later than p2's """
		t1 = p1[RESPONSE].getLoadTime()
		t2 = p2[RESPONSE].getLoadTime()
		if t1 < t2:
			return -1
		elif t1 == t2:
			return 0
		else:
			return 1

	def socketSortOrder(self, p1, p2):
		""" Returns -1, 0, or 1 depending on whether p1's request was received earlier, equal to or later than p2's """
		s1 = p1[REQUEST].getIPandPort()
		s2 = p2[REQUEST].getIPandPort()
		t1 = p1[REQUEST].getCreationTime()
		t2 = p2[REQUEST].getCreationTime()
		if s1 < s2:
			return -1
		elif s1 == s2:
			if t1 < t2:
				return -1
			elif t1 == t2:
				return 0
			else:
				return 1
		else:
			return 1

	def findMessage(self, objectid):
		""" Returns the request or response matching the objectid """
		p = self.findMessagePair(objectid)
		if p != None:
			if id(p[REQUEST]) == objectid:
				return p[REQUEST]
			elif id(p[RESPONSE]) == objectid:
				return p[RESPONSE]
		return None

	def findMessagePair(self, objectid):
		""" Returns the message containing a req or resp matching the objectid """
		for p in self.messagelist:
			if id(p[REQUEST]) == objectid or id(p[RESPONSE]) == objectid:
				return p
		return None

	def createMLHeader(self):
		""" Returns the message list HTML header """
		if self.sortorder == REQUEST_SORT_ORDER:
			sortoptions = '[Time Request Received] <a href="/WEBDEBUG/self.showMessageList(\'RESPONSE\')">[Time Response Received]</a> <a href="/WEBDEBUG/self.showMessageList(\'SOCKET\')">[Socket Creation Time]</a>'
		elif self.sortorder == RESPONSE_SORT_ORDER:
			sortoptions = '[Time Response Received] <a href="/WEBDEBUG/self.showMessageList(\'REQUEST\')">[Time Request Received]</a> <a href="/WEBDEBUG/self.showMessageList(\'SOCKET\')">[Socket Creation Time]</a>'
		elif self.sortorder == SOCKET_SORT_ORDER:
			sortoptions = '[Socket Creation Time] <a href="/WEBDEBUG/self.showMessageList(\'REQUEST\')">[Time Request Received]</a> <a href="/WEBDEBUG/self.showMessageList(\'RESPONSE\')">[Time Response Received]</a>'
		else:
			sortoptions = 'INVALID SORT ORDER'
		return 	"""
			<h3>Contents of Message List</h3><p>
			Sorted on %s <p>
			%s message pairs<br>First pair added %s<br>Last pair added %s<p>
			""" %  (sortoptions,
				len(self.messagelist),
				time.ctime(self.firstmessagetime),
				time.ctime(self.lastmessagetime))

	def toSpreadsheet(self):
		""" Converts the contents of the list to CSV, suitable for importing into a spreadsheet """
		if len(self.messagelist) == 0:
			return "<h3>Message List Empty</h3>"
		csv = "%s%s%s\n" % (HTTPrequest.SPREADSHEET_HEADER, SEPARATOR, HTTPresponse.SPREADSHEET_HEADER)
		# render in chronological order all message pairs to CSV
		for p in self.messagelist:
			csv = csv + "%s%s%s\n" % (p[REQUEST].toSpreadsheet(), SEPARATOR, p[RESPONSE].toSpreadsheet())
		return csv

	def toHTML(self):
		""" Converts the contents of the list to HTML, suitable for display in a browser """
		if len(self.messagelist) == 0:
			return "<h3>Message List Empty</h3>"
		html = self.createMLHeader()
		html = html + "<table border=0>"
		counter = 1
		totrawreqsize = 0
		totrawrespsize = 0
		totthroughput = 0
		statistics = { }
		sockets = { }
		# render in chronological order all message pairs to HTML
		for p in self.messagelist:
			if p[RESPONSE] == None:
				responseHTML = "<b>Response not yet received from server</b>"
			else:
				responseHTML = p[RESPONSE].toHTML()
			html = html + """
			<tr><th align=left><font color=red>HTTP browser request %s</font></th></tr><tr><td>%s</td></tr>
			<tr><th align=left><font color=red>HTTP server response %s</font></th></tr><tr><td>%s</td></tr>
			<tr><td><hr></td></tr>
			<tr><td>&nbsp;</td></tr>
			""" %  (counter, p[REQUEST].toHTML(),
				counter, responseHTML)
			totrawreqsize = totrawreqsize + p[REQUEST].getRawSize()
			if p[RESPONSE] != None:
				totrawrespsize = totrawrespsize + p[RESPONSE].getRawSize()
			if p[RESPONSE] != None:
				totthroughput = totthroughput + p[RESPONSE].getThroughput()
			counter = counter + 1
			self.calcStatistics(statistics, p[RESPONSE])
			self.trackSockets(sockets, p[REQUEST].getIPandPort())
		html = html + """
		</table><p>
		<table border=0>
		<tr><td><b>Total request/response pairs:</b></td><td>%d</td></tr>
		<tr><td><b>Total size of requests:</b></td><td>%s bytes (average of %d bytes/request)</td></tr>
		<tr><td><b>Total size of responses:</b></td><td>%s bytes (average of %d bytes/response)</td></tr>
		<tr><td><b>Total size of both:</b></td><td>%s bytes</td></tr>
		<tr><td><b>Average response throughput:</b></td><td>%d bytes/sec</td></tr>
		<tr><td><b>Total number of sockets used:</b></td><td>%d</td></tr>
		</table><p>
		""" %  (len(self.messagelist),
			totrawreqsize, totrawreqsize/len(self.messagelist),
			totrawrespsize,  totrawrespsize/len(self.messagelist),
			totrawreqsize + totrawrespsize,
			totthroughput/len(self.messagelist),
			len(sockets))
		html = html + self.statisticsToHTML(statistics)
		return html

	def statisticsToHTML(self, stats):
		""" generates the content-type statistics in HTML """
		if stats == { }:
			return ""
		html = "<table border=1><tr><th colspan=3>Content-type Statistics</th></tr>"
		html = html + "<tr><th>Content-type</th><th>Number</th><th>Total size (bytes)</th></tr>\n"
		keys = stats.keys()
		keys.sort()
		for c in keys:
			html = html + '<tr><td>%s</td><td align="right">%s</td><td align="right">%s</td></tr>\n' % (c, stats[c][STAT_NUMB], stats[c][STAT_LEN])
		html = html + "</table><p>"
		return html

	def calcStatistics(self, stats, response):
		""" Keeps track of content statistics """
		if response == None:
			return
		type = response.getContentType()
		length = response.getContentLength()
		if length == 0:
			return
		if type == None:
			self.addToStats(stats, "UNKNOWN", length)
		else:
			self.addToStats(stats, type, length)

	def addToStats(self, stats, type, length):
		if stats.get(type) == None:
			stats[type] = [1, length]
		else:
			stats[type][STAT_NUMB] = stats[type][STAT_NUMB] + 1
			stats[type][STAT_LEN] = stats[type][STAT_LEN] + length

	def trackSockets(self, socket_map, ip_and_port):
		""" Adds the string ip_and_port to the socket_map, if it's already in the map, we just increment the count """
		if socket_map.get(ip_and_port) == None:
			socket_map[ip_and_port] = 1
		else:
			socket_map[ip_and_port] = socket_map[ip_and_port] + 1

myList = MessageList()
