"""
WebDebug TCP connection handler

Handles requests made to the webdebug server.
Responsibilities:
- Instantiating HTTPrequest and HTTPresponse objects
- Passing them to MessageList
- Handling all WebDebug commands


WebDebug is a web development proxy server that captures and displays HTTP
requests and responses.
Copyright (C) 1999  Paul Clip.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""

### IMPORTS

from WDconst import *
import HTTPrequest
import HTTPresponse
import MessageList
import WDconfig

import SocketServer
from socket import *
import string
import traceback
import time
import cPickle


### CONSTANTS

WD_HTTP_HEADER = "HTTP/1.0 200 OK\r\nContent-type: text/html\r\nContent-length: %s\r\n\r\n"

ACTION = "action"

WD_MENU = """
<html><title>WebDebug Menu</title>
<body bgcolor="#ffffff">
<font face="arial" size="+1">WebDebug Menu</font>
<p>
<table border=0>
<tr><td>
	<a href="/WEBDEBUG/self.showMessageList()">Show Message List</a>
</td></tr>
<tr><td>
	<a href="/WEBDEBUG/self.clearMessageList()">Clear Message List</a>
</td></tr>
<tr><td>
	<form action="/WEBDEBUG/self.paramAction()" method="GET">
	Load Message List: <input type="text" name="filename">
	<input type=submit value="Load">
	<input type="hidden" name="action" value="self.loadMessageList">
	</form>
</td></tr>
<tr><td>
	<form action="/WEBDEBUG/self.paramAction()" method="GET">
	Save Message List: <input type="text" name="filename">
	<input type=submit value="Save">
	<input type="hidden" name="action" value="self.saveMessageList">
	</form>
</td></tr>
<tr><td>
	<form action="/WEBDEBUG/self.paramAction()" method="GET">
	Save stats in Tab-delimited format: <input type="text" name="filename">
	<input type=submit value="Save">
	<input type="hidden" name="action" value="self.saveTabbedMessageList">
	</form>
</td></tr>
</table><p>
%s
</body></html>
"""

WD_TEMPLATE = """
<html><title>WebDebug</title>
<body bgcolor="#ffffff">
<a href="/WEBDEBUG/self.displayWebDebugMenu()">
	<font face="arial" size="+1">WebDebug Menu</font>
</a>
<p>
%s
<p>
<a href="/WEBDEBUG/self.displayWebDebugMenu()">
	<font face="arial" size="+1">WebDebug Menu</font>
</a>
</body></html>
"""

### Classes

class WDhandler(SocketServer.StreamRequestHandler):

	def handle(self):
		""" Handles the HTTP request: read request, send it to destination, read response, add both to MessageList """
		doKeepAlive = YES
		self.socketOpen = NO
		while doKeepAlive:
			self.httpRequest = HTTPrequest.HTTPrequest(self.rfile, self.client_address)
			if self.isWebDebugCommand():
				self.execWebDebugCommand()
				doKeepAlive = NO
			else:
				print "HANDLING %s" % self.httpRequest.getPath()
				MessageList.myList.addRequest(self.httpRequest)
				self.getResponse()
				# add request and response to session manager
				MessageList.myList.addResponse(self.httpRequest, self.httpResponse)
				self.httpResponse.sendResponse(self.wfile)
				print "\tDONE %s" % self.httpRequest.getPath()
				doKeepAlive = self.useKeepAlive()

	def getResponse(self):
		""" Connects to a remote host, submits the request and creates the response """
		if not self.socketOpen:
			self.s = socket(AF_INET, SOCK_STREAM)
			self.s.connect(self.httpRequest.getHost(), self.httpRequest.getPort())
			self.outsock = self.s.makefile("wb")
			self.insock = self.s.makefile("rb")
			self.socketOpen = YES
		self.httpRequest.sendRequest(self.outsock)
		self.httpResponse = HTTPresponse.HTTPresponse(self.insock, self.s.getpeername())
		if not self.useKeepAlive():
			self.outsock.close()
			self.insock.close()
			self.s.close()
			self.socketOpen = NO


	def useKeepAlive(self):
		""" Returns true if we should use Keep-Alive, false otherwise """
		return WDconfig.KEEP_ALIVE and self.httpRequest.isKeepAlive() and self.httpResponse.isKeepAlive()

	def isWebDebugCommand(self):
		""" Returns true if the requested path is in fact a WebDebug command """
		if startsWith(self.httpRequest.getPath(), WD_PREFIX):
			return TRUE
		else:
			return FALSE

	def execWebDebugCommand(self):
		""" Executes the WebDebug command encoded in the URL path """
		path = self.httpRequest.getPath()
		if len(path) == len(WD_PREFIX):
			self.displayWebDebugMenu()
		else:
			try:
				commandpath = path[len(WD_PREFIX):]
				if string.find(commandpath, '?') != -1:
					# if the URL's path has arguments, we don't want to evaluate them
					commandpath = commandpath[:string.find(commandpath, '?')]
				eval(commandpath)
			except Exception, e:
				error = "Couldn't execute %s<p>%s" % (self.httpRequest.getPath(), str(e))
				self.displayWebDebugError(error, e)

	def displayWebDebugMenu(self, optional_message=""):
		""" Returns a list of options to the browser """
		self.sendWebDebugResponse(WD_MENU % optional_message)

	def displayWebDebugError(self, str, exception):
		""" Returns an error to the browser """
		self.sendWebDebugResponse("Error: %s" % str)
		traceback.print_exc()

	def sendWebDebugResponse(self, str):
		""" Send a response to the browser """
		response = WD_HTTP_HEADER % len(str)
		response = response + str
		self.wfile.write(response)
		self.wfile.flush()

	def paramAction(self):
		""" Executes a WebDebug browser command with parameters given """
		contents = self.httpRequest.getContents()
		action = contents.get(ACTION)
		if action == None:
			self.displayWebDebugMenu("<b>Error: no action supplied</b>")
		else:
			params = ""
			for k in contents.keys():
				if k != ACTION:
					params = params + "%s='%s'," % (k, contents.get(k))
			command = "%s(%s)" % (action, params[:-1])
			eval(command)

	def clearMessageList(self):
		MessageList.myList.clearList()
		self.displayWebDebugMenu("<b>Message List cleared</b>")

	def loadMessageList(self, filename):
		""" Loads a message list from disk """
		try:
			f = open(filename, 'r')
			MessageList.myList = cPickle.load(f)
			f.close()
			self.displayWebDebugMenu("<b>Message List loaded</b>")
		except:
			self.displayWebDebugMenu("<b>Could not load file: %s</b>" % filename)

	def saveMessageList(self, filename):
		""" Saves a message list to disk """
		try:
			f = open(filename, 'w')
			cPickle.dump(MessageList.myList, f)
			f.close()
			self.displayWebDebugMenu("<b>Message List saved</b>")
		except:
			self.displayWebDebugMenu("<b>Could not save file: %s</b>" % filename)

	def saveTabbedMessageList(self, filename):
		""" Saves a tab-delimited message list to disk """
		try:
			f = open(filename, 'w')
			csv = MessageList.myList.toSpreadsheet()
			f.write(csv)
			f.close()
			self.displayWebDebugMenu("<b>Tab-delimited Message List saved</b><p><pre>%s</pre>" % csv)
		except:
			self.displayWebDebugMenu("<b>Could not save tab-delimited file: %s</b>" % filename)

	def showMessageList(self, sort_order = CURRENT_SORT_ORDER):
		""" Displays the contents of the message list """
		MessageList.myList.sortOn(sort_order)
		html = MessageList.myList.toHTML()
		self.sendWebDebugResponse(WD_TEMPLATE % html)

	def displayResponse(self, objectid, randomfoo, message):
		""" Returns a specific response to the browser """
		# randomfoo is used to make the browser believe each URL is unique
		# otherwise it may return cached responses if they have the same objectid
		r = MessageList.myList.findMessage(objectid)
		if r != None:
			if message == 'all':
				r.sendResponse(self.wfile)
			elif message == 'html':
				html = r.getRawContents()
				html = string.replace(html, '<', '&lt;')
				html = string.replace(html, '>', '&gt;')
				self.sendWebDebugResponse("<pre>%s</pre>" % html)
			else:
				self.sendWebDebugResponse("Couldn't execute " + self.httpRequest.getPath())
		else:
			self.sendWebDebugResponse("Couldn't find " + self.httpRequest.getPath())
